// ESFrameObserverSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.simtoolsgui.GUISwarm;
import swarm.simtoolsgui.GUISwarmImpl;

import swarm.analysis.EZGraph;
import swarm.analysis.EZGraphImpl;

import swarm.collections.ListImpl;

import java.util.*;


/**
 * The Observer contains the Model and the graphic widgets
 * @author  Pietro Terna and several of his students
 */
public class ESFrameObserverSwarm extends GUISwarmImpl {

    /** update frequency */
    public int displayFrequency;

    /** the time at which we stop the simulation (if != 0) */
    public int timeToFinish;

    /** displaying all messages on the console */
    public boolean verboseChoice;

    /** two ActionGroup for sequence of GUI events */
    public ActionGroup displayActions1, displayActions2;

    /** the single Schedule instance */
    public Schedule displaySchedule;

    /** the Swarm we are observing */
    public ESFrameModelSwarm eSFrameModelSwarm;

    /** our graphics or EZGraph output to files*/
    public EZGraphImpl 
	totalTimeLengthGraph=null;

    /** the address of the probe, if we need to access it; remember casting
     *  (ESFrameObserverSwarm$1$ESFrameObserverProbeMap)
     */
    Object observerProbe;

    /** Constructor for class */
    public ESFrameObserverSwarm (Zone aZone) {

	super(aZone);

	// fill in the relevant parameters (only two, in this case).
	displayFrequency = 1;
	verboseChoice=false;
	timeToFinish=0;

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  ESFrameObserverSwarm$1$ESFrameObserverProbeMap.class
        class ESFrameObserverProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, ESFrameObserverSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, ESFrameObserverSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }

	    public ESFrameObserverProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("displayFrequency");
		addVar ("verboseChoice");
		addVar ("timeToFinish");
	    }
	}
	// Install our custom probeMap class directly into the
	// probeLibrary
	observerProbe = new ESFrameObserverProbeMap (aZone, getClass ());
	Globals.env.probeLibrary.setProbeMap$For
	    ((ESFrameObserverProbeMap) observerProbe, getClass ());
    }

    /** Create the objects used to display the model. */
    public Object buildObjects () {

	super.buildObjects ();
	
	// First, we create the model that we're actually observing. The
	// model is a subswarm of the observer.

	// creating the ESFrameModelSwarm via lispAppArchivers, which
	// refers to the jesframe.scm file where we can modify parameters
	// in a stable way without compiling again the code
	eSFrameModelSwarm =
	    (ESFrameModelSwarm)Globals.env.lispAppArchiver.getWithZone$key(
						      getZone(), "eSFrameModelSwarm");
	// instead of using the direct constructor way
	// eSFrameModelSwarm = new ESFrameModelSwarm (getZone ());

	// Now create probe objects on the model and ourselves. This gives a
	// simple user interface to let the user change parameters.

	Globals.env.createArchivedProbeDisplay (eSFrameModelSwarm,
						"eSFrameModelSwarm");
	Globals.env.createArchivedProbeDisplay (this, "eSFrameObserverSwarm");

	// Instruct the control panel to wait for a button event: we
	// halt here until someone hits a control panel button so the
	// user can get a chance to fill in parameters before the
	// simulation runs
	getControlPanel ().setStateStopped ();

	// to change the verboseChoice via probe you have to wright 'true'
	// or 'false'; not simply 't' or 'f'
	// setting the 'verbose' value in StartESFrame
	StartESFrame.verbose=verboseChoice;

	// First, let the model swarm build its objects.
	eSFrameModelSwarm.buildObjects ();

	// Create the graph widget to display the total production time
	// for all the finished orders divided by their total recipe lengths

	totalTimeLengthGraph = new EZGraphImpl
	    (getZone (), "Ratio total time / total lengths", "Time",
	     "Ratio", "totalTimeLengthGraph");

	Globals.env.setWindowGeometryRecordName
	    (totalTimeLengthGraph, "totalTimeLengthGraph");

	// the data for the totalTimeLengthGraph
	totalTimeLengthGraph.createSequence$withFeedFrom$andSelector
	    ("ratio", eSFrameModelSwarm,
	     SwarmUtils.getSelector(eSFrameModelSwarm,"getTimeLengthRatio"));

	return this;
    }
    

  /**
   * Create the actions necessary for the simulation. This is where
   * the schedule is built (but not run!)  Here we create a display
   * schedule - this is used to display the state of the world and
   * check for user input. 
   */

    public Object buildActions () {
	int i;

	super.buildActions();

	// First, let our model swarm build its own schedule.
	eSFrameModelSwarm.buildActions();

	// Create two ActionGroup for display
	displayActions1 = new ActionGroupImpl (getZone());
	displayActions2 = new ActionGroupImpl (getZone());

	// Add the methods to the ActionGroup

	// Schedule the update of the graphic widgets
	// EZGraph files
	displayActions1.createActionTo$message
	    (totalTimeLengthGraph,
	     SwarmUtils.getSelector(totalTimeLengthGraph, "step"));

	// Schedule the update of the probe displays
	displayActions2.createActionTo$message
	    (Globals.env.probeDisplayManager,
	     SwarmUtils.getSelector(Globals.env.probeDisplayManager,"update"));

	// Schedule an update for the whole user
	// interface code.  This is crucial: without this, no
	// graphics update and the control panel will be
	// dead. It's best to put it at the end of the display
	// schedule
	displayActions2.createActionTo$message(getActionCache(),
       		      SwarmUtils.getSelector(getActionCache(),"doTkEvents"));

	// check to finish
	displayActions2.createActionTo$message(this,
       		      SwarmUtils.getSelector(this,"finish"));

	// The display schedule. Note the repeat interval: display
	// is frequently the slowest part of a simulation, so redrawing less
	// frequently can be a help.
	displaySchedule = new ScheduleImpl (getZone (), displayFrequency);

	// insert ActionGroup instance on the repeating Schedule
	// instance

	for (i=0;i<displayFrequency;i++)
	    {
		// displayActions1 is activated only in the last tick of each
		// unit of time (a day, a shift, ...), when some conclusive
		// statistics have to be displayed
		if(i==displayFrequency-1)
		    displaySchedule.at$createAction (i, displayActions1);
		
		displaySchedule.at$createAction (i, displayActions2);
	    }
	
	return this;
    }

    /**
       activateIn: - activate the schedules so they're ready to run.
       The swarmContext argument has to do with what we were activated
       *in*.  Typically the ObserverSwarm is the top-level Swarm, so
       it's activated in "null". But other Swarms and Schedules and
       such will be activated inside of us.  */

    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves (just pass along the context).
	super.activateIn (swarmContext);

	// Activate the model swarm in ourselves. The model swarm is a
	// subswarm of the observer swarm.
	eSFrameModelSwarm.activateIn (this);

	// Now activate our schedule in ourselves. This arranges for
	// the execution of the schedule we built.
	displaySchedule.activateIn (this);

	// Activate returns the swarm activity - the thing that's ready to run.
	return getActivity();
  }

    /** drop the Observer, but after having dropped the EZGraph */
    public void drop() {

	System.out.println("\nPress newly Quit to exit.");
	getControlPanel ().setStateStopped ();

	if(totalTimeLengthGraph!=null)
	    totalTimeLengthGraph.drop();
	totalTimeLengthGraph=null;

	super.drop();
    }

    /** to finish automatically, if timeToFinish!=0 */
    public void finish(){
	if(timeToFinish != 0 && timeToFinish<=Globals.env.getCurrentTime())
	    {
		System.out.println("jES virtually quitted at time " + 
				   Globals.env.getCurrentTime());
		getControlPanel ().setStateQuit ();;
	    }
    }

}
